 /*
   Unix ttcp-wifi header file.
   Copyright (C) Alamy Liu <alamyliu@yahoo.com> 2002

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/

#ifndef	_TTCP_H_
#define	_TTCP_H_

#include	<stdio.h>

#include	<sys/time.h>		// for getrusage()
#include	<sys/resource.h>
#include	<unistd.h>


#define	MAX_LOOP_TIMES		1000


/*====================================================================*/
#if defined(SunOS53)
	#include <sys/rusage.h>
	int		getrusage(int who, struct rusage *rusage);
	/*
	 * The following #defines address a VERY useful feature the people of
	 * SunSoft introduced in SunOS5.3: nanoseconds. They called their
	 * struct timespec - and as far as I obseverved, they only use it in
	 * struct rusage. The fucking thing: all other call's use the old
	 * timeval with microseconds, so nothing is compatible anymore if you
	 * do measuring with gettimeofday and getrusage. I resolved this
	 * pitty, with using timespec allover and adjusting the results of
	 * gettimeofday vi NANO_ADJ.  Happy new Year, Elmar Bartel,
	 * bartel@informatik.tu-muenchen.de 29.12.1994 It's getting better
	 * all the time: they renamed it to "tv_nsec" but deliver only values
	 * from 0..10^6-1 -- or say: I haven't observed other values. Ok, we
	 * are prepared, but till now we define SKIPOVER to 1000000. When
	 * SunSoft decides to deliver real ns, simply define SKIPOVER to
	 * 1000000000.  Despite this, have a happy new Year, Elmar Bartel,
	 * bartel@informatik.tu-muenchen.de 30.12.1994 
	 */
#define TIMEVAL			timespec
#define	SKIPOVER		1000000
#define FRACT			tv_nsec
#define FRACT_TO_SEC(tv)	(((double)tv.tv_nsec)/SKIPOVER)
#define FRACT_TO_DECI(tv)	(tv.tv_nsec/(SKIPOVER/10))
#define FRACT_TO_CENT(tv)	(tv.tv_nsec/(SKIPOVER/100))
#define FRACT_TO_MILL(tv)	(tv.tv_nsec/(SKIPOVER/1000))
#define	NANO_ADJ(tv)		(tv).tv_nsec*= (SKIPOVER/1000000)
#else							/* SunOS53 */
#define TIMEVAL			timeval
#define	SKIPOVER		1000000
#define FRACT			tv_usec
#define FRACT_TO_SEC(tv)	(((double)tv.tv_usec)/SKIPOVER)
#define FRACT_TO_DECI(tv)	(tv.tv_usec/(SKIPOVER/10))
#define FRACT_TO_CENT(tv)	(tv.tv_usec/(SKIPOVER/100))
#define FRACT_TO_MILL(tv)	(tv.tv_usec/(SKIPOVER/1000))
#define	NANO_ADJ(tv)
#endif							/* SunOS53 */

#define	TimeVal			struct TIMEVAL

/*====================================================================*/
#if !defined(RUSAGE_SELF)
	/*
	 * definitions for systems without getrusage(2) calls, but times(2): +
	 * SystemV 4.01 from AT&T + SunOS2.[45] (SunOS2.3 has it!!)
	 */
#include <sys/times.h>
#define	RUSAGE_SELF	0
#define	RUSAGE_CHILDREN	1
struct rusage {
	struct timeval  ru_utime;	/* user time used */
	struct timeval  ru_stime;	/* system time used */
	int             ru_maxrss;	/* maximum resident set size */
	int             ru_ixrss;	/* currently 0 */
	int             ru_idrss;	/* integral resident set size */
	int             ru_isrss;	/* currently 0 */
	int             ru_minflt;	/* page faults not requiring physical I/O */
	int             ru_majflt;	/* page faults requiring physical I/O */
	int             ru_nswap;	/* swaps */
	int             ru_inblock;	/* block input operations */
	int             ru_oublock;	/* block output operations */
	int             ru_msgsnd;	/* messages sent */
	int             ru_msgrcv;	/* messages received */
	int             ru_nsignals;	/* signals received */
	int             ru_nvcsw;	/* voluntary context switches */
	int             ru_nivcsw;	/* involuntary context switches */
};

#endif							/* RUSAGE_SELF */
/*====================================================================*/


/*
 * this describes the connection(s) to our remote site(s)
 */
typedef struct {
	char           *HostName;
	char           *IPName;
	int             Socket;
	FILE           *fin;
	FILE           *fout;
} RemoteConnection;

typedef	struct {
	TimeVal			time;		/* Time at which timing started */
	struct rusage	ru;			/* rusage at start */
} Timer;

/*
 * this structure holding testing data
 */
typedef struct {
	unsigned long	nRecvBytes;	// the amount of received bytes
	unsigned long	nSendBytes;	// the amount of transfered bytes
	unsigned long	nBuffer;	// the number of buffers transferd/received
	unsigned long	nFailed;	// number of failed comparisions
	double			cput;		// cpu time for transmission in s
	double			realt;		// real time for transmission in s
} BenchmarkData;

/*
 * this structure is filled from the commandline options 
 */
typedef struct {
	int             PidCalls;	/* do PidCalls to getpid() */
	int             udp;		/* ==1: use UDP instead of TCP */
	int             Compare;	/* compare the received data */
	int             StreamPattern;	/* transmit a pattern stream */
	int             Transmit;	/* ==1: transmitting, ==0 receiveing */
	int             Loop;		/* loop times of measurement */
	int             b_flag;		/* ==1: use mread() */
	int             SockOpt;	/* Debug-SocketOptions */
	int             NoDelay;	/* ==1: set TCP_NODELAY socket option */
	int             BufCnt;		/* number of buffers to send */
	int             BufLen;		/* buffer length */
	int             FixedDataSize;	/* if>0: calc BufCnt/BufLen from * *
									 * this */
	int             Window,		/* ==1: Set TCP send/receive window size */
	                SndWin,		/* TCP send window size */
	                RcvWin;		/* TCP receive window sizes */
	int             Verbose;	/* ==1: more diagnostics */
	char           *MulticastChannel;	/* we send multicast traffic */
	short           MulticastPort;
	int             GapLength;	/* Gap length between packets */
	int             inetd;		/* run in inetd mode */
	short           Service;	/* the service-port inetd listens */
	char           *OutputFmt;	/* the output format */
	char           *RemHost;	/* the remote host to connect to */
	char           *InitString;	/* the stream init string */
	int             RemoteNumber;	/* */
	int             Wifi;		/* Wifi testing case (1..3) */
} Options;

typedef	struct {
	char *		buf;			/* the buffer to read from, to write to */
	char *		ExpectBuf;		/* when comparing the data */
	BenchmarkData *	benchmarkArray;
				/* [Options.Loop] space to save measure data */
} MemoryResource;


void	StartTimer(Timer *timer);
void	StopTimer(Timer *timer, double *cput, double *realt);


void	SigPipe(int dummy);

void	Transfer_data(int fd, Options *opt);
void	Receive_data(int fd, Options *opt);

#endif	// _TTCP_H_

